@@ -1,10 +1,10 @@
 package Dancer2::Serializer::YAML;
 # ABSTRACT: Serializer for handling YAML data
-$Dancer2::Serializer::YAML::VERSION = '0.166001';
+$Dancer2::Serializer::YAML::VERSION = '0.300004';
 use Moo;
 use Carp 'croak';
 use Encode;
-use Class::Load 'load_class';
+use Module::Runtime 'use_module';
 
 with 'Dancer2::Core::Role::Serializer';
 
@@ -16,7 +16,7 @@ sub from_yaml { __PACKAGE__->deserialize(@_) }
 sub to_yaml { __PACKAGE__->serialize(@_) }
 
 # class definition
-sub BUILD { load_class('YAML') }
+sub BUILD { use_module('YAML') }
 
 sub serialize {
     my ( $self, $entity ) = @_;
@@ -42,7 +42,7 @@ Dancer2::Serializer::YAML - Serializer for handling YAML data
 
 =head1 VERSION
 
-version 0.166001
+version 0.300004
 
 =head1 DESCRIPTION
 
@@ -77,15 +77,13 @@ This is an helper available to transform a Perl data structure to YAML.
 
 Calling this function will B<not> trigger the serialization's hooks.
 
-=head1 METHODS
-
 =head1 AUTHOR
 
 Dancer Core Developers
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2015 by Alexis Sukrieh.
+This software is copyright (c) 2020 by Alexis Sukrieh.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
