@@ -1,6 +1,6 @@
 package Dancer2::Session::Simple;
 # ABSTRACT: in-memory session backend for Dancer2
-$Dancer2::Session::Simple::VERSION = '0.166001';
+$Dancer2::Session::Simple::VERSION = '0.300004';
 use Moo;
 use Dancer2::Core::Types;
 use Carp;
@@ -25,6 +25,13 @@ sub _retrieve {
     return $s;
 }
 
+sub _change_id {
+    my ( $class, $old_id, $new_id ) = @_;
+
+    $SESSIONS->{$new_id} = $class->_retrieve($old_id);
+    delete $SESSIONS->{$old_id};
+}
+
 sub _destroy {
     my ( $class, $id ) = @_;
     delete $SESSIONS->{$id};
@@ -49,7 +56,7 @@ Dancer2::Session::Simple - in-memory session backend for Dancer2
 
 =head1 VERSION
 
-version 0.166001
+version 0.300004
 
 =head1 DESCRIPTION
 
@@ -78,7 +85,7 @@ Dancer Core Developers
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2015 by Alexis Sukrieh.
+This software is copyright (c) 2020 by Alexis Sukrieh.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
